#!/bin/python3
# SPDX-License-Identifier: GPL-2.0-or-later

import pathlib
import subprocess
import time
from typing import Optional

MAX_SSHOT_CAPTURE_TIME_S = 10

class GeeqieScreenshot:
    def __init__(self, sshot_dir: Optional[pathlib.Path] = None):
        """Initializes GeeqieScreenshot

        May raise ImportError.
        """
        # We run this here to make it easier to use this as an optional, best-effort dependency
        import psutil

        if sshot_dir:
            self._sshot_dir = sshot_dir
        else:
            self._sshot_dir = pathlib.Path("/tmp/gq_test_screenshots/")

    def capture(self, base_filename: str, xvfb_run_process: subprocess.Popen) -> Optional[pathlib.Path]:
        import psutil

        self._sshot_dir.mkdir(parents=True, exist_ok=True)
        escaped_name = base_filename.replace('.', '_')
        sshot_name = f"{int(time.time())}_{escaped_name}.png"
        sshot_file = self._sshot_dir / sshot_name

        # We borrow the geeqie process's environment, to pick up the appropriate DISPLAY and
        # XAUTHORITY values from xvfb.
        geeqie_env = None

        xvfb_process = psutil.Process(xvfb_run_process.pid)
        for child in xvfb_process.children():
            if not child.name().endswith("geeqie"):
                continue

            # Found the geeqie process.  Now find the environment variables we care about.
            geeqie_env = child.environ()
            break

        if not geeqie_env:
            return None

        # Attempt to take the screenshot
        subprocess.run(args=["import", "-window", "root", str(sshot_file)], timeout=MAX_SSHOT_CAPTURE_TIME_S, env=geeqie_env)
        return sshot_file



#!/bin/python3
# SPDX-License-Identifier: GPL-2.0-or-later

import itertools
import optparse
import os
import pathlib
import shutil
import subprocess
import sys
import tempfile
import time

DEBUG_MODE=False


def start_xephyr(xargs: list) -> list[subprocess.Popen, str]:
    """Starts a Xephyr instance with the specified arguments.

    Args:
        xargs: A list of arguments to include.  Key/value pairs should be separated with
          an "=" character (which will be split into two arguments internally).  Because
          Xephyr handles repeated arguments, arguments will _not_ be de-duplicated.

          A small set of default arguments will be used, but may be overridden by the
          specified arguments.

    Returns:
        List containing the subprocess.Popen handle for the Xephyr process, as well as the
        DISPLAY for the server.
    """

    # TODO: use pass_fds with the -displayfd argument to avoid blind sleep
    # Argument merging behavior:
    default_xephyr_args = [
            ["-ac"], ["-br"], ["-reset"], ["-terminate", "2"], ["-screen", "1920x1080"]]

    xephyr_args = [arg.split("=", 2) for arg in xargs]
    xephyr_arg_keys = set([arg[0] for arg in xephyr_args])

    # Prepend default args to xephyr_args only-if they haven't been overridden.
    accepted_defaults = [arg for arg in default_xephyr_args
                         if arg[0] not in xephyr_arg_keys]

    # Concatenate and flatten final args list.
    xephyr_args = itertools.chain.from_iterable(accepted_defaults + xephyr_args)

    xephyr_display = ":55"
    xephyr_cmd = ["Xephyr", *xephyr_args, xephyr_display]

    if DEBUG_MODE:
        print("Xephyr: " + repr(xephyr_cmd))

    xephyr_proc = subprocess.Popen(xephyr_cmd, stderr=subprocess.DEVNULL)
    time.sleep(1)

    return [xephyr_proc, xephyr_display]


def add_symlink(parent_dir: pathlib.Path, spec: str):
    """Creates a symlink in the specified location.

    Args:
        parent_dir:
            The directory under which symlinks will be created.  Paths outside this
            directory will not be modified (including through prior symlinks).
    """
    dest_str, target_str = spec.split("=", 2)
    dest_path = (parent_dir / dest_str).resolve()
    target_path = pathlib.Path(target_str)
    # Validate target path.
    target_path.resolve(strict=True)

    if not dest_path.is_relative_to(parent_dir):
        raise RuntimeError(f"Symlink dest spec {spec} resolved {dest_str} to {dest_path}, "
                           f"which is not a child of {parent_dir}")

    if DEBUG_MODE:
        print(f"Symlink dest spec {spec} resolved {dest_str} to {dest_path}, which is a "
              f"child of {parent_dir}")

    dest_path.parent.mkdir(parents=True, exist_ok=True)
    dest_path.symlink_to(target_path)


# Inspired by isolate-test.sh, but intended for interactive development and diagnostics
def main(args, passthrough_argv) -> int:
    # We need to clean this up by hand, because there's a race condition with dbus
    # infrastructure shutting down asynchronously, so it often takes two attempts.
    test_home = tempfile.TemporaryDirectory(prefix="gq_xephyr_")
    xephyr_proc = None
    try:
        env = {}

        env["HOME"] = pathlib.Path(test_home.name)
        # Change to the temp homedir.
        os.chdir(env["HOME"])

        env["XDG_CONFIG_HOME"] = env["HOME"] / ".config"
        env["XDG_CONFIG_HOME"].mkdir(parents=True)

        # Mode setting as required by the spec.
        # https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
        # (Mode only affects the leaf dir; not parents)
        env["XDG_RUNTIME_DIR"] = env["HOME"] / ".runtime"
        env["XDG_RUNTIME_DIR"].mkdir(mode=0o700, parents=True)

        # This file must exist to prevent a modal dialog in main.cc being triggered.
        gq_config = env["XDG_CONFIG_HOME"] / "geeqie"
        gq_config.mkdir()
        (gq_config / "accels.ini").touch()

        sandbox_bin = env["HOME"] / "bin"
        sandbox_bin.mkdir()
        env["PATH"] = f"{sandbox_bin}:/usr/bin:/bin"

        for symlink_spec in args.link_path:
            add_symlink(env["HOME"], symlink_spec)

        # If `bwrap` is installed, symlink it into the sandbox PATH. Used by glycin GdkPixuf.
        bwrap_path = shutil.which("bwrap")
        if bwrap_path:
            add_symlink(env["HOME"], f"bin/bwrap={bwrap_path}")

        xephyr_proc, xephyr_display = start_xephyr(args.xarg)
        env["DISPLAY"] = xephyr_display

        str_env = {k: str(v) for k, v in env.items()}
        print("Variables in isolated environment:", file=sys.stderr)
        subprocess.run(["dbus-run-session", "--", "env"], env=str_env, stdout=sys.stderr)
        print("", file=sys.stderr)

        # TODO: Use single dbus-run-session?
        # Optionally starts a window manager.
        # We don't clean it up manually; it should exit when the Xephyr server dies.
        if args.wm:
            wm_proc = subprocess.Popen(args.wm, env=str_env, stdout=sys.stderr)
            time.sleep(1)

        run_cmd = ["dbus-run-session", "--", *passthrough_argv]
        print(f"running: {repr(run_cmd)}", file=sys.stderr)
        subprocess.run(run_cmd, env=str_env)

    finally:
        if xephyr_proc and xephyr_proc.poll() is None:
            print("Terminating Xephyr…", file=sys.stderr)
            xephyr_proc.terminate()

        try:
            test_home.cleanup()
        except (ConnectionAbortedError, OSError):
            # Cleanup race condition with dbus infrastructure
            print("First cleanup attempt failed; sleeping and retrying…", file=sys.stderr)
            time.sleep(2)

            test_home.cleanup()

    return 0

if __name__ == "__main__":
    # Using optparse instead of argparse to correctly handle Xephyr args that start with
    # a dash.  See https://github.com/python/cpython/issues/53580
    parser = optparse.OptionParser()
    parser.add_option("--xarg", action="append", default=[],
                        help="Arg for Xephyr server; repeatable. Use key=value when needed.")
    parser.add_option("--debug", action="store_true", help="Enables harness debug logging")
    parser.add_option("--link_path", action="append", default=[],
                        help=("File or dir to symlink into the isolated homedir.  Use "
                              "dest=source, with dest relative to isolated homedir.  Parent "
                              "directories will be created."))
    parser.add_option("--wm",
                        help="A window-manager to start.")
    args, passthrough_argv = parser.parse_args()

    DEBUG_MODE = args.debug
    if DEBUG_MODE:
        print(f"sa: {repr(args)}\npa: {repr(passthrough_argv)}\n")

    exit(main(args, passthrough_argv))



#!/bin/python3
#**********************************************************************
# Copyright (C) 2023 - The Geeqie Team
#
# Author: Colin Clark
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#**********************************************************************

import os
import pathlib
import re
import subprocess
import sys
import time
import traceback
from typing import Optional

MAX_GEEQIE_INIT_TIME_S = 10
MAX_REMOTE_CMD_TIME_S = 10
MAX_GEEQIE_SHUTDOWN_TIME_S = 10


class GeeqieTestError(Exception):
    """A custom exception type that forwards the geeqie exit code.

    Attributes:
        exit_code: An int with the exit code to return when image-test.py exits.
    """

    def __init__(self, description: str, geeqie_proc: Optional[subprocess.Popen]):
        """GeeqieTestError constructor.

        Args:
            description: A description of the failure state.
            geeqie_proc: The Popen wrapper for the geeqie process.  The exception
                message will be generic if this is not provided.
        """
        self._description = description
        if geeqie_proc is not None:
            self._geeqie_proc_provided = True
            self._proc_exit_code = geeqie_proc.poll()
        else:
            self._geeqie_proc_provided = False
            self._proc_exit_code = 1

        if self._proc_exit_code:
            self.exit_code = self._proc_exit_code
        else:
            self.exit_code = 1

    def __str__(self) -> str:
        if self._geeqie_proc_provided:
            if self._proc_exit_code is None:
                # geeqie process is still running.
                verb = "stalled"
            else:
                # geeqie process exited unexpectedly.
                verb = "crashed"

            return f"geeqie {verb} during: {self._description}"

        # Generic error.  This will automatically be preceded by the Exception
        # class name, "GeeqieTestError", so no extra decoration is needed.
        return self._description


class GeeqieScriptedTest:
    def __init__(self, argv):
        self.geeqie_exe = argv[1]
        self.test_image_dir = pathlib.Path(argv[2])
        if not self.test_image_dir.is_dir():
            raise RuntimeError(
                    f"test_image_dir is not an existing directory: {self.test_image_dir}")

        self.container_dir = pathlib.Path.home() / "container_dir"
        self.symlink_images_dir = self.container_dir / "images_dir"

    def run(self, script: list[tuple[str, Optional[re.Pattern]]]) -> None:
        # First off, we set up our directory structure, which will be an
        # otherwise-empty directory, containing a symlink to (test_image_dir).
        # During the test, geeqie will start in the container dir, then will switch
        # to the (test_image_dir) symlink, then will switch back to the container
        # dir.
        #
        # The expectation is that all FDs will be unreffed and freed by the time
        # geeqie exits.

        # TODO: Validate whether or not Geeqie showed any leaked FDs.

        self.container_dir.mkdir()
        self.symlink_images_dir.symlink_to(self.test_image_dir)

        if not self.symlink_images_dir.is_dir():
            raise RuntimeError("symlinking images dir failed")

        # All geeqie commands start with this.
        geeqie_cmd_prefix = ["xvfb-run", "--auto-servernum", self.geeqie_exe]

        # Start geeqie and have it start in the container directory
        # TODO(xsdg): Note that killing the `xvfb-run` script will not forward that
        # signal to the geeqie process.  See
        # <https://unix.stackexchange.com/questions/291804/howto-terminate-xvfb-run-properly>.
        # So we should modify or replace xvfb-run to allow us to kill an errant geeqie process.
        geeqie_proc = subprocess.Popen(
                args=[*geeqie_cmd_prefix, f"--file={self.container_dir}"])

        # try/finally to ensure we clean up geeqie_proc
        try:
            start_time = time.monotonic()
            # Wait up to MAX_GEEQIE_INIT_TIME_S for remote to initialize.
            command_fifo_path = pathlib.Path(os.environ["XDG_CONFIG_HOME"], "geeqie/layouts")

            while not command_fifo_path.exists():
                time.sleep(1)
                if time.monotonic() - start_time > MAX_GEEQIE_INIT_TIME_S:
                    raise GeeqieTestError("init (before creating command fifo)", geeqie_proc)

            # We make sure Geeqie can stay alive for 2 seconds after initialization.
            time.sleep(2)
            if geeqie_proc.poll() is not None:
                raise GeeqieTestError("2-second post-init check", geeqie_proc)

            # Run script (checking for crashes between each command)
            commands = [item[0] for item in script]
            exp_results = [item[1] for item in script]
            cmd_results = []

            for command in commands:
                cmd_results.append(subprocess.run(
                        args=[*geeqie_cmd_prefix, command],
                        capture_output=True, text=True, timeout=MAX_REMOTE_CMD_TIME_S))

                # Check if Geeqie crashed (which would cause xvfb-run to terminate)
                time.sleep(1)
                if geeqie_proc.poll() is not None:
                    raise GeeqieTestError(f"remote command processing: {command}", geeqie_proc)

            # Request shutdown
            subprocess.run(args=[*geeqie_cmd_prefix, "--quit"],
                           timeout=MAX_REMOTE_CMD_TIME_S)

            # If there's a timeout, or the exit code isn't zero, flag it.
            try:
                if geeqie_proc.wait(MAX_GEEQIE_SHUTDOWN_TIME_S) != 0:
                    raise GeeqieTestError("shutdown", geeqie_proc)
            except subprocess.TimeoutExpired:
                # Re-raise as shutdown error.
                raise GeeqieTestError("shutdown", geeqie_proc)

            # Validate results.
            for exp_pat, result in zip(exp_results, cmd_results):
                if not exp_pat:
                    continue

                if not exp_pat.match(result.stdout):
                    raise GeeqieTestError(
                            f"Unexpected output for command '{result.args[-1]}': "
                            f"{repr(result.stdout)} did not match {repr(exp_pat)}",
                            geeqie_proc=None)

        except subprocess.TimeoutExpired as e:
            # This means one of the remote commands failed.  Re-raise as test error,
            # and include the original exception (to show which remote failed).
            raise GeeqieTestError("remote command timeout", geeqie_proc=None) from e

        finally:
            if geeqie_proc.poll() is None:
                # Stash the true error now, before we alter the state of geeqie_proc.
                exit_error = GeeqieTestError("shutdown", geeqie_proc)

                # geeqie is still running after we requested it to exit, so try SIGTERM.
                geeqie_proc.terminate()

                try:
                    geeqie_proc.wait(timeout=15)
                except subprocess.TimeoutExpired:
                    print("sending SIGKILL to geeqie")
                    geeqie_proc.kill()

                raise exit_error
